#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Colour.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/Obj.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/SpotLight.h"
#include <QString>


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  // now set the inital camera values
  m_cameraIndex=0;
  m_moveMode=MOVEEYE;
  m_drawHelp=true;
  m_fov=65.0;
  m_aspect=1024.0/768.0;
  m_projection=ngl::PERSPECTIVE;
  m_rotation=0;
  startTimer(25);
  m_lights[0]=true;
  m_lights[1]=true;
  m_lights[2]=true;

}
void GLWindow::createCameras()
{
  // create a load of cameras
  ngl::Camera cam;
  ngl::Camera Tcam;
  ngl::Camera Scam;
  ngl::Camera Fcam;
  // set the different vectors for the camera positions
  ngl::Vector EYE(0,5,8);
  ngl::Vector LOOK=0.0;
  ngl::Vector UP(0,1,0);

  ngl::Vector TEYE(0,8,0);
  ngl::Vector TLOOK=0.0;
  ngl::Vector TUP(0,0,1);

  ngl::Vector SEYE(8,0,0);
  ngl::Vector SLOOK=0.0;
  ngl::Vector SUP(0,1,0);

  ngl::Vector FEYE(5,2,8);
  ngl::Vector FLOOK(0.0,0.0,0.0);
  ngl::Vector FUP(0,1,0);
  // finally set the cameras shape and position
  cam.set(EYE,LOOK,UP);
  cam.setShape(m_fov,m_aspect, 0.2f,150.0f,m_projection);
  cam.setOrthoParams(-1,1,-1,1,0.1,10.0);
  m_cameras.push_back(cam);
  Tcam.set(TEYE,TLOOK,TUP);
  Tcam.setShape(m_fov,m_aspect, 0.5f,150.0f,m_projection);
  Tcam.setOrthoParams(-0.1,0.1,-0.1,0.1,0.1,10.0);
  m_cameras.push_back(Tcam);
  Scam.set(SEYE,SLOOK,SUP);
  Scam.setShape(m_fov,m_aspect, 0.5f,150.0f,m_projection);
  Scam.setOrthoParams(-0.1,0.1,-0.1,0.1,0.1,10.0);
  m_cameras.push_back(Scam);
  Fcam.set(FEYE,FLOOK,FUP);
  Fcam.setShape(m_fov,m_aspect, 0.5f,150.0f,m_projection);
  Fcam.setOrthoParams(-0.1,0.1,-0.1,0.1,0.1,10.0);
  m_cameras.push_back(Fcam);
}
// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
  glEnable(GL_DEPTH_TEST);
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  #ifdef WIN32
  	glewInit();
  #endif
  Init->initGlew();
  createCameras();
  // create an instance of the VBO primitive
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // create a plane
  prim->createVBOQuadPlane("plane",20,20,30,30,ngl::Vector(0,1,0));
  prim->createVBOSphere("sphere",1.0,150);

  // now to load the shader and set the values
  // grab an instance of shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // load a frag and vert shaders for colour no shading
  shader->loadShader("Colour","shaders/Colour.vs","shaders/Colour.fs");
  // set this as the active shader
  shader->useShader("Colour");
  shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
	shader->setShaderParam4f("Colour","Colour",1,1,1,1);
  // load a frag and vert shaders
  shader->loadShader("gl3xTest","shaders/Vertex.vs","shaders/Fragment.fs");
  // set this as the active shader
  shader->useShader("gl3xTest");
  // now pass the modelView and projection values to the shader
  shader->setShaderParamFromMatrix("gl3xTest","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  shader->setShaderParam1i("gl3xTest","Normalize",true);
  shader->setShaderParam1i("gl3xTest","numLightsEnabled",3);
  shader->setShaderParam3f("gl3xTest","eye",m_cameras[m_cameraIndex].getEye()[0],m_cameras[m_cameraIndex].getEye()[1],m_cameras[m_cameraIndex].getEye()[2]);

  // now set the material and light values
  ngl::Material m(ngl::GOLD);
  m.use();
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  // set the viewport
  glViewport(0,0,_w,_h);
  // calculate the aspect ratio
  m_aspect=(float)_w/_h;
  // now set the camera values
  setCameraShape();
 }

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // Rotation based on the mouse position for our global
  // transform
  ngl::Transformation trans;
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);
  // now set this value in the shader for the current Camera, we do all 3 elements as they
  // can all change per frame
  shader->useShader("gl3xTest");
  shader->setShaderParamFromMatrix("gl3xTest","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  shader->setShaderParam4f("gl3xTest","eye",m_cameras[m_cameraIndex].getEye()[0],m_cameras[m_cameraIndex].getEye()[1],m_cameras[m_cameraIndex].getEye()[2],1.0);
  int numEnabledLights=0;
  ngl::Light L1(ngl::Vector(-1,1,0),ngl::Colour(1,1,1,1),ngl::LIGHTREMOTE);
  ngl::Light L2(ngl::Vector(0,1,0),ngl::Colour(0.5,0.5,0.5,1),ngl::LIGHTLOCAL);
  ngl::Light L3(ngl::Vector(1,1,0),ngl::Colour(1,1,1,1),ngl::LIGHTREMOTE);
  L1.setSpecColour(ngl::Colour(0.5,0.5,0.5));
  L2.setSpecColour(ngl::Colour(0.5,0.5,0.5));
  L3.setSpecColour(ngl::Colour(0.5,0.5,0.5));

  if( m_lights[0] == true)
  {
    L1.enable();
    numEnabledLights++;
  }

  if( m_lights[1] == true)
  {
    L2.setAttenuation(0,0.2,0);
    L2.enable();
    numEnabledLights++;

  }
  if( m_lights[2] == true)
  {
    L3.enable();
    numEnabledLights++;
  }
  shader->setShaderParam1i("gl3xTest","numLightsEnabled",numEnabledLights);

  // get the VBO instance and draw the built in teapot
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();

  m_transformStack.getCurrentTransform().setRotation(0,m_rotation,0);

  shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());

  prim->draw("teapot");
  // now we are going to draw the grid in wireframe mode
  glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  // load the colour  shader and set the values,
  shader->useShader("Colour");
  // at the outer level of the drawing we set the view and projection
  shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cameras[m_cameraIndex].getModelView());
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  // now push the tx stack it helps to use { } to block off the transforms
  m_transformStack.pushTransform();
  {
    // translate the grid down by -0.5 so it is at the base level of the objects
    m_transformStack.getCurrentTransform().setPosition(0,-0.5,0);
    // now we pass this modelling transform to the shader ModelMatrix
    shader->setShaderParamFromMatrix("Colour","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
    // finally draw the plane
    prim->draw("plane");
  }
  m_transformStack.popTransform();
  // now go back to solid drawing
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  // activate the shader
  shader->useShader("gl3xTest");
  // push a transform
  m_transformStack.pushTransform();
  {
    // translate into position
    m_transformStack.getCurrentTransform().setPosition(-1.8,0.1,0);
    // rotate object
    m_transformStack.getCurrentTransform().setRotation(m_rotation,m_rotation,m_rotation);
    // pass values to the shader
    shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
    // draw
    prim->draw("cube");
  }
  m_transformStack.popTransform();
  // now we draw the football
  m_transformStack.pushTransform();
  {
    // set the transforms for pos and scale
    m_transformStack.getCurrentTransform().setPosition(1.5,0,0);
    m_transformStack.getCurrentTransform().setScale(0.5,0.5,0.5);
    m_transformStack.getCurrentTransform().setRotation(m_rotation,m_rotation,m_rotation);

    // pass to the shader
    shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
    // draw the football
    prim->draw("football");
  }
  m_transformStack.popTransform();

  // now we are going to draw the help
  if (m_drawHelp==true)
  {
   // first we need to fallback to the fixed functionality GL pipeline
   shader->useNullProgram();
   // create a font for drawing
   QFont font;
   // set it to bold
   font.setBold(true);
   // an make it 16 point
   font.setPointSize(16);
   // now we use the deprecated glColour command to set the colour for the text
   // this will not work with 3/4 core profile but neither will Qt at present
   glColor3f(1,1,1);
   // now render the text using the QT renderText helper function
   renderText(10,18,"Use Arrow Keys to move Camera i and o to move in and out",font);
   renderText(10,18*2,"Use keys 0-4 to switch cameras",font);
   renderText(10,18*3,"r roll, y yaw, p pitch",font);
   renderText(10,18*4,"e move eye mode",font);
   renderText(10,18*5,"l move look mode",font);
   renderText(10,18*6,"b move eye and look mode",font);
   renderText(10,18*7,"s slide mode",font);
   renderText(10,18*8,"+/- change fov of Camera",font);
   renderText(10,18*9,"5 ortho projection 6 perspective projection",font);
   renderText(10,18*10,"z,x,c toggle lights on and off",font);
   renderText(10,18*11,"h toggles help",font);
   glColor3f(1,1,0);
   // now we are going to construct a string to draw the mode information
   QString mode;
   switch ( m_moveMode)
   {
     case MOVEEYE : mode=QString("Move Eye"); break;
     case MOVELOOK : mode=QString("Move Look"); break;
     case MOVEBOTH : mode=QString("Move Both"); break;
     case MOVESLIDE : mode=QString("Move Slide"); break;
   }
   QString text=QString("Active Camera %1 current mode=%2").arg(m_cameraIndex).arg(mode);
   renderText(600,18,text,font);
   text=QString("FOV = %1 ").arg(m_fov);
   renderText(600,18*2,text,font);
   renderText(600,18*3,"ModelView Matrix",font);
   // now we use the QString sprintf
   ngl::Matrix m=m_cameras[m_cameraIndex].getModelView();
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[0],m.openGL()[1],m.openGL()[2],m.openGL()[3]);
   renderText(600,18*4,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[4],m.openGL()[5],m.openGL()[6],m.openGL()[7]);
   renderText(600,18*5,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[8],m.openGL()[9],m.openGL()[10],m.openGL()[11]);
   renderText(600,18*6,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[12],m.openGL()[13],m.openGL()[14],m.openGL()[15]);
   renderText(600,18*7,text,font);


   renderText(600,18*8,"Projection Matrix",font);
   // now we use the QString sprintf
   m=m_cameras[m_cameraIndex].getProjection();
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[0],m.openGL()[1],m.openGL()[2],m.openGL()[3]);
   renderText(600,18*9,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[4],m.openGL()[5],m.openGL()[6],m.openGL()[7]);
   renderText(600,18*10,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[8],m.openGL()[9],m.openGL()[10],m.openGL()[11]);
   renderText(600,18*11,text,font);
   text.sprintf("[ %+0.4f %+0.4f %+0.4f %+0.4f]",m.openGL()[12],m.openGL()[13],m.openGL()[14],m.openGL()[15]);
   renderText(600,18*12,text,font);
   text=QString("Light 0 = %1 Light 1 %2 Light 2 %3").arg(m_lights[0] ? "On" : "Off").arg(m_lights[1] ? "On" : "Off").arg(m_lights[2] ? "On" : "Off");
   renderText(600,18*13,text,font);

  }

}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}


void GLWindow::setCameraShape()
{
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  int cameraSize=m_cameras.size();
  for(int i=0; i<cameraSize; ++i)
  {
    m_cameras[i].setShape(m_fov,m_aspect, 0.2f,150.0f,m_projection);
    m_cameras[i].setOrthoParams(-1,1,-1,1,1,10.0);
  }
  shader->useShader("gl3xTest");
  shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
  shader->useShader("Colour");
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cameras[m_cameraIndex].getProjection());
}

void GLWindow::keyPressEvent(
                               QKeyEvent *_event
                              )
{
  const static float keyIncrement=0.2;
  // this method is called every time the main window recives a key event.
  // we then switch on the key value and set the camera in the GLWindow
  // find which key and do appropriate event
  switch (_event->key())
  {
  case Qt::Key_Escape : { QApplication::exit(EXIT_SUCCESS); break; }
  case Qt::Key_1 : { m_cameraIndex=0; break; }
  case Qt::Key_2 : { m_cameraIndex=1; break; }
  case Qt::Key_3 : { m_cameraIndex=2; break; }
  case Qt::Key_4 : { m_cameraIndex=3; break; }

  case Qt::Key_E : { m_moveMode=MOVEEYE; break; }
  case Qt::Key_L : { m_moveMode=MOVELOOK; break; }
  case Qt::Key_B : { m_moveMode=MOVEBOTH; break; }
  case Qt::Key_S : { m_moveMode=MOVESLIDE; break; }
  case Qt::Key_H : { m_drawHelp^=true; break; }
  case Qt::Key_5 : { m_projection=ngl::ORTHOGRAPHIC; setCameraShape(); break; }
  case Qt::Key_6 : { m_projection=ngl::PERSPECTIVE; setCameraShape(); break; }
  case Qt::Key_Plus : { ++m_fov; setCameraShape(); break; }
  case Qt::Key_Minus :{ --m_fov; setCameraShape(); break; }
  case Qt::Key_R : { m_cameras[m_cameraIndex].roll(3);  break; }
  case Qt::Key_P : { m_cameras[m_cameraIndex].pitch(3); break; }
  case Qt::Key_Y : { m_cameras[m_cameraIndex].yaw(3);   break; }
  case Qt::Key_Z : { m_lights[0]^=true; break; }
  case Qt::Key_X : { m_lights[1]^=true; break; }
  case Qt::Key_C : { m_lights[2]^=true; break; }

  case Qt::Key_Left :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(keyIncrement,0,0);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(keyIncrement,0,0); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(keyIncrement,0,0); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(keyIncrement,0,0);    break;}
    }
  break;
  } // end move left
  case Qt::Key_Right :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(-keyIncrement,0,0);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(-keyIncrement,0,0); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(-keyIncrement,0,0); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(-keyIncrement,0,0);    break;}
    }
  break;
  } // end move right
  case Qt::Key_Up :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(0,keyIncrement,0);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(0,keyIncrement,0); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(0,keyIncrement,0); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(0,keyIncrement,0);    break;}
    }
  break;
  } // end move up
  case Qt::Key_Down :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(0,-keyIncrement,0);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(0,-keyIncrement,0); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(0,-keyIncrement,0); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(0,-keyIncrement,0);    break;}
    }
  break;
  } // end move down
  case Qt::Key_O:
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(0,0,keyIncrement);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(0,0,keyIncrement); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(0,0,keyIncrement); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(0,0,keyIncrement);    break;}
    }
  break;
  } // end move out
  case Qt::Key_I :
  {
    switch (m_moveMode)
    {
      case MOVEEYE :   { m_cameras[m_cameraIndex].moveEye(0,0,-keyIncrement);  break;}
      case MOVELOOK :  { m_cameras[m_cameraIndex].moveLook(0,0,-keyIncrement); break;}
      case MOVEBOTH :  { m_cameras[m_cameraIndex].moveBoth(0,0,-keyIncrement); break;}
      case MOVESLIDE : { m_cameras[m_cameraIndex].slide(0,0,-keyIncrement);    break;}
    }
  break;
  } // end move in
  default : break;
  }
  // finally update the GLWindow and re-draw
 updateGL();
}

void GLWindow::timerEvent(QTimerEvent *_e)
{
  ++m_rotation;
  updateGL();
}
